/* (c) Magnus Auvinen. See licence.txt in the root of the distribution for more information. */
/* If you are missing that file, acquire a complete release at teeworlds.com.                */
#include <stdio.h>

#include <engine/editor.h>
#include <engine/engine.h>
#include <engine/friends.h>
#include <engine/graphics.h>
#include <engine/textrender.h>
#include <engine/demo.h>
#include <engine/map.h>
#include <engine/storage.h>
#include <engine/sound.h>
#include <engine/serverbrowser.h>
#include <engine/shared/demo.h>
#include <engine/shared/config.h>

#include <game/generated/protocol.h>
#include <game/generated/client_data.h>

#include <game/localization.h>
#include <game/version.h>
#include "render.h"
#include "teecomp.h"

#include "gameclient.h"

#include "components/binds.h"
#include "components/broadcast.h"
#include "components/camera.h"
#include "components/chat.h"
#include "components/console.h"
#include "components/controls.h"
#include "components/countryflags.h"
#include "components/damageind.h"
#include "components/debughud.h"
#include "components/effects.h"
#include "components/emoticon.h"
#include "components/flow.h"
#include "components/hud.h"
#include "components/items.h"
#include "components/killmessages.h"
#include "components/mapimages.h"
#include "components/maplayers.h"
#include "components/menus.h"
#include "components/motd.h"
#include "components/particles.h"
#include "components/players.h"
#include "components/nameplates.h"
#include "components/scoreboard.h"
#include "components/skins.h"
#include "components/sounds.h"
#include "components/spectator.h"
#include "components/timemessages.h"
#include "components/voting.h"
#include "components/race_demo.h"
#include "components/ghost.h"
#include "components/teecomp_stats.h"

CGameClient g_GameClient;

// instanciate all systems
static CKillMessages gs_KillMessages;
static CTimeMessages gs_TimeMessages;
static CCamera gs_Camera;
static CChat gs_Chat;
static CMotd gs_Motd;
static CBroadcast gs_Broadcast;
static CGameConsole gs_GameConsole;
static CBinds gs_Binds;
static CParticles gs_Particles;
static CMenus gs_Menus;
static CSkins gs_Skins;
static CCountryFlags gs_CountryFlags;
static CFlow gs_Flow;
static CHud gs_Hud;
static CDebugHud gs_DebugHud;
static CControls gs_Controls;
static CEffects gs_Effects;
static CScoreboard gs_Scoreboard;
static CSounds gs_Sounds;
static CEmoticon gs_Emoticon;
static CDamageInd gsDamageInd;
static CVoting gs_Voting;
static CSpectator gs_Spectator;
static CRaceDemo gs_RaceDemo;
static CGhost gs_Ghost;
static CTeecompStats gs_TeecompStats;

static CPlayers gs_Players;
static CNamePlates gs_NamePlates;
static CItems gs_Items;
static CMapImages gs_MapImages;

static CMapLayers gs_MapLayersBackGround(CMapLayers::TYPE_BACKGROUND);
static CMapLayers gs_MapLayersForeGround(CMapLayers::TYPE_FOREGROUND);

CGameClient::CStack::CStack() { m_Num = 0; }
void CGameClient::CStack::Add(class CComponent *pComponent) { m_paComponents[m_Num++] = pComponent; }

const char *CGameClient::Version() { return GAME_VERSION; }
const char *CGameClient::NetVersion() { return GAME_NETVERSION; }
const char *CGameClient::GetItemName(int Type) { return m_NetObjHandler.GetObjName(Type); }

// char MapName[32] = {0};

int CGameClient::GetCountryIndex(int Code)
{
	int Index = g_GameClient.m_pCountryFlags->Find(Code);
	if(Index < 0)
	{
		Index = g_GameClient.m_pCountryFlags->Find(-1);
		if(Index < 0)
			Index = 0;
	}	
	return Index;
}

void CGameClient::OnConsoleInit()
{
	m_pEngine = Kernel()->RequestInterface<IEngine>();
	m_pClient = Kernel()->RequestInterface<IClient>();
	m_pGraphics = Kernel()->RequestInterface<IGraphics>();
	m_pTextRender = Kernel()->RequestInterface<ITextRender>();
	m_pSound = Kernel()->RequestInterface<ISound>();
	m_pInput = Kernel()->RequestInterface<IInput>();
	m_pConsole = Kernel()->RequestInterface<IConsole>();
	m_pStorage = Kernel()->RequestInterface<IStorage>();
	m_pDemoPlayer = Kernel()->RequestInterface<IDemoPlayer>();
	m_pDemoRecorder = Kernel()->RequestInterface<IDemoRecorder>();
	m_pServerBrowser = Kernel()->RequestInterface<IServerBrowser>();
	m_pEditor = Kernel()->RequestInterface<IEditor>();
	m_pFriends = Kernel()->RequestInterface<IFriends>();
	
	// setup pointers
	m_pBinds = &::gs_Binds;
	m_pGameConsole = &::gs_GameConsole;
	m_pParticles = &::gs_Particles;
	m_pMenus = &::gs_Menus;
	m_pSkins = &::gs_Skins;
	m_pCountryFlags = &::gs_CountryFlags;
	m_pChat = &::gs_Chat;
	m_pFlow = &::gs_Flow;
	m_pCamera = &::gs_Camera;
	m_pControls = &::gs_Controls;
	m_pEffects = &::gs_Effects;
	m_pSounds = &::gs_Sounds;
	m_pMotd = &::gs_Motd;
	m_pDamageind = &::gsDamageInd;
	m_pMapimages = &::gs_MapImages;
	m_pVoting = &::gs_Voting;
	m_pHud = &::gs_Hud;
	m_pRaceDemo = &::gs_RaceDemo;
	m_pTeecompStats = &::gs_TeecompStats;
	m_pScoreboard = &::gs_Scoreboard;
	m_pGhost = &::gs_Ghost;
	m_pItems = &::gs_Items;
	
	// make a list of all the systems, make sure to add them in the corrent render order
	m_All.Add(m_pSkins);
	m_All.Add(m_pCountryFlags);
	m_All.Add(m_pMapimages);
	m_All.Add(m_pEffects); // doesn't render anything, just updates effects
	m_All.Add(m_pParticles);
	m_All.Add(m_pBinds);
	m_All.Add(m_pControls);
	m_All.Add(m_pCamera);
	m_All.Add(m_pSounds);
	m_All.Add(m_pVoting);
	m_All.Add(m_pParticles); // doesn't render anything, just updates all the particles
	m_All.Add(m_pRaceDemo);
	
	m_All.Add(&gs_MapLayersBackGround); // first to render
	m_All.Add(&m_pParticles->m_RenderTrail);
	m_All.Add(m_pItems);
	m_All.Add(&gs_Players);
	m_All.Add(m_pGhost);
	m_All.Add(&gs_MapLayersForeGround);
	m_All.Add(&m_pParticles->m_RenderExplosions);
	m_All.Add(&gs_NamePlates);
	m_All.Add(&m_pParticles->m_RenderGeneral);
	m_All.Add(m_pDamageind);
	m_All.Add(&gs_Hud);
	m_All.Add(&gs_Spectator);
	m_All.Add(&gs_Emoticon);
	m_All.Add(&gs_KillMessages);
	m_All.Add(&gs_TimeMessages);
	m_All.Add(m_pChat);
	m_All.Add(&gs_Broadcast);
	m_All.Add(&gs_DebugHud);
	m_All.Add(&gs_Scoreboard);
	m_All.Add(m_pRaceDemo);
	m_All.Add(m_pTeecompStats);
	m_All.Add(m_pMotd);
	m_All.Add(m_pMenus);
	m_All.Add(m_pGameConsole);
	
	// build the input stack
	m_Input.Add(&m_pMenus->m_Binder); // this will take over all input when we want to bind a key
	m_Input.Add(&m_pBinds->m_SpecialBinds);
	m_Input.Add(m_pGameConsole);
	m_Input.Add(m_pChat); // chat has higher prio due to tha you can quit it by pressing esc
	m_Input.Add(m_pMotd); // for pressing esc to remove it
	m_Input.Add(m_pMenus);
	m_Input.Add(&gs_Spectator);
	m_Input.Add(&gs_Emoticon);
	m_Input.Add(m_pControls);
	m_Input.Add(m_pBinds);
	
	// add the some console commands
	Console()->Register("team", "i", CFGFLAG_CLIENT, ConTeam, this, "Switch team");
	Console()->Register("kill", "", CFGFLAG_CLIENT, ConKill, this, "Kill yourself");
	Console()->Register("+minimap", "", CFGFLAG_CLIENT, ConMinimap, this, "Show the minimap"); // Gamer: minimap
	Console()->Register("send_position", "", CFGFLAG_CLIENT, ConSendPosition, this, "Send your current relative position via teamchat");
	
	// register server dummy commands for tab completion
	Console()->Register("tune", "si", CFGFLAG_SERVER, 0, 0, "Tune variable to value");
	Console()->Register("tune_reset", "", CFGFLAG_SERVER, 0, 0, "Reset tuning");
	Console()->Register("tune_dump", "", CFGFLAG_SERVER, 0, 0, "Dump tuning");
	Console()->Register("change_map", "?r", CFGFLAG_SERVER, 0, 0, "Change map");
	Console()->Register("restart", "?i", CFGFLAG_SERVER, 0, 0, "Restart in x seconds");
	Console()->Register("broadcast", "r", CFGFLAG_SERVER, 0, 0, "Broadcast message");
	Console()->Register("say", "r", CFGFLAG_SERVER, 0, 0, "Say in chat");
	Console()->Register("set_team", "ii", CFGFLAG_SERVER, 0, 0, "Set team of player to team");
	Console()->Register("set_team_all", "i", CFGFLAG_SERVER, 0, 0, "Set team of all players to team");
	Console()->Register("add_vote", "sr", CFGFLAG_SERVER, 0, 0, "Add a voting option");
	Console()->Register("remove_vote", "s", CFGFLAG_SERVER, 0, 0, "remove a voting option");
	Console()->Register("force_vote", "ss?r", CFGFLAG_SERVER, 0, 0, "Force a voting option");
	Console()->Register("clear_votes", "", CFGFLAG_SERVER, 0, 0, "Clears the voting options");
	Console()->Register("vote", "r", CFGFLAG_SERVER, 0, 0, "Force a vote to yes/no");


	// propagate pointers
	m_UI.SetGraphics(Graphics(), TextRender());
	m_RenderTools.m_pGraphics = Graphics();
	m_RenderTools.m_pUI = UI();
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->m_pClient = this;
	
	// let all the other components register their console commands
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnConsoleInit();
	
	
	//
	Console()->Chain("player_name", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_clan", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_country", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_use_custom_color", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_color_body", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_color_feet", ConchainSpecialInfoupdate, this);
	Console()->Chain("player_skin", ConchainSpecialInfoupdate, this);
	
	//
	m_SuppressEvents = false;
}

void CGameClient::OnInit()
{
	int64 Start = time_get();

	// set the language
	g_Localization.Load(g_Config.m_ClLanguagefile, Storage(), Console());

	// TODO: this should be different
	// setup item sizes
	for(int i = 0; i < NUM_NETOBJTYPES; i++)
		Client()->SnapSetStaticsize(i, m_NetObjHandler.GetObjSize(i));

	// load default font	
	static CFont *pDefaultFont = 0;
	char aFilename[512];
	IOHANDLE File = Storage()->OpenFile("fonts/DejaVuSans.ttf", IOFLAG_READ, IStorage::TYPE_ALL, aFilename, sizeof(aFilename));
	if(File)
	{
		io_close(File);
		pDefaultFont = TextRender()->LoadFont(aFilename);
		TextRender()->SetDefaultFont(pDefaultFont);
	}
	if(!pDefaultFont)
		Console()->Print(IConsole::OUTPUT_LEVEL_STANDARD, "gameclient", "failed to load font. filename='fonts/DejaVuSans.ttf'");
	
	// init all components
	for(int i = m_All.m_Num-1; i >= 0; --i)
		m_All.m_paComponents[i]->OnInit();

	// setup load amount// load textures
	for(int i = 0; i < g_pData->m_NumImages; i++)
	{
		g_pData->m_aImages[i].m_Id = Graphics()->LoadTexture(g_pData->m_aImages[i].m_pFilename, IStorage::TYPE_ALL, CImageInfo::FORMAT_AUTO, 0);
		g_GameClient.m_pMenus->RenderLoading();
	}

	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnReset();
	
	int64 End = time_get();
	char aBuf[256];
	str_format(aBuf, sizeof(aBuf), "initialisation finished after %.2fms", ((End-Start)*1000)/(float)time_freq());
	Console()->Print(IConsole::OUTPUT_LEVEL_DEBUG, "gameclient", aBuf);
	
	m_ServerMode = SERVERMODE_PURE;

	m_IsRace = false;
	m_RaceMsgSent = false;
	m_ShowOthers = -1;
	for(int i = 0; i < 2; i++)
		m_aFlagPos[i] = vec2(-1, -1);
		
	// Teecomp grayscale flags
	Graphics()->UnloadTexture(g_pData->m_aImages[IMAGE_GAME_GRAY].m_Id); // Already loaded with full color, unload
	g_pData->m_aImages[IMAGE_GAME_GRAY].m_Id = -1;

	CImageInfo Info;
	if(!Graphics()->LoadPNG(&Info, g_pData->m_aImages[IMAGE_GAME_GRAY].m_pFilename, IStorage::TYPE_ALL))
		return;

	unsigned char *d = (unsigned char *)Info.m_pData;
	int Step = Info.m_Format == CImageInfo::FORMAT_RGBA ? 4 : 3;

	for(int i=0; i < Info.m_Width*Info.m_Height; i++)
	{
		int v = (d[i*Step]+d[i*Step+1]+d[i*Step+2])/3;
		d[i*Step] = v;
		d[i*Step+1] = v;
		d[i*Step+2] = v;
	}

	int aFreq[256];
	int OrgWeight;
	int NewWeight;
	int FlagX = 384;
	int FlagY = 256;
	int FlagW = 128;
	int FlagH = 256;
	int Pitch = Info.m_Width*4;

	for(int f=0; f<2; f++)
	{
		OrgWeight = 0;
		NewWeight = 192;
		for(int i=0; i<256; i++)
			aFreq[i] = 0;

		// find most common frequence
		for(int y=FlagY; y<FlagY+FlagH; y++)
			for(int x=FlagX+FlagW*f; x<FlagX+FlagW*(1+f); x++)
			{
				if(d[y*Pitch+x*4+3] > 128)
					aFreq[d[y*Pitch+x*4]]++;
			}
		
		for(int i = 1; i < 256; i++)
		{
			if(aFreq[OrgWeight] < aFreq[i])
				OrgWeight = i;
		}

		// reorder
		int InvOrgWeight = 255-OrgWeight;
		int InvNewWeight = 255-NewWeight;
		for(int y=FlagY; y<FlagY+FlagH; y++)
			for(int x=FlagX+FlagW*f; x<FlagX+FlagW*(1+f); x++)
			{
				int v = d[y*Pitch+x*4];
				if(v <= OrgWeight*1.25f) // modified for contrast
					v = (int)(((v/(float)OrgWeight) * NewWeight));
				else
					v = (int)(((v-OrgWeight)/(float)InvOrgWeight)*InvNewWeight + NewWeight);
				d[y*Pitch+x*4] = v;
				d[y*Pitch+x*4+1] = v;
				d[y*Pitch+x*4+2] = v;
			}
	}

	g_pData->m_aImages[IMAGE_GAME_GRAY].m_Id = Graphics()->LoadTextureRaw(Info.m_Width, Info.m_Height, Info.m_Format, Info.m_pData, Info.m_Format, 0);
	mem_free(Info.m_pData);
}

void CGameClient::DispatchInput()
{
	// handle mouse movement
	float x = 0.0f, y = 0.0f;
	Input()->MouseRelative(&x, &y);
	if(x != 0.0f || y != 0.0f)
	{
		for(int h = 0; h < m_Input.m_Num; h++)
		{
			if(m_Input.m_paComponents[h]->OnMouseMove(x, y))
				break;
		}
	}
	
	// handle key presses
	for(int i = 0; i < Input()->NumEvents(); i++)
	{
		IInput::CEvent e = Input()->GetEvent(i);
		
		for(int h = 0; h < m_Input.m_Num; h++)
		{
			if(m_Input.m_paComponents[h]->OnInput(e))
			{
				//dbg_msg("", "%d char=%d key=%d flags=%d", h, e.ch, e.key, e.flags);
				break;
			}
		}
	}
	
	// clear all events for this frame
	Input()->ClearEvents();	
}


int CGameClient::OnSnapInput(int *pData)
{
	return m_pControls->SnapInput(pData);
}

void CGameClient::OnConnected()
{
	m_Layers.Init(Kernel());
	m_Collision.Init(Layers());
	
	RenderTools()->RenderTilemapGenerateSkip(Layers());

	for(int i = 0; i < m_All.m_Num; i++)
	{
		m_All.m_paComponents[i]->OnMapLoad();
		m_All.m_paComponents[i]->OnReset();
	}
	
	CServerInfo CurrentServerInfo;
	Client()->GetServerInfo(&CurrentServerInfo);
	
	m_ServerMode = SERVERMODE_PURE;
	m_LastSendInfo = 0;
	
	g_Config.m_GfxSpecZoom = 100;
	
	// send the inital info
	SendInfo(true);

	m_LastGameOver = 0;
	m_LastRoundStartTick = 0;
	m_aLastFlagCarrier[0] = -1;
	m_aLastFlagCarrier[1] = -1;
 	
	for(int i = 0; i < 2; i++)
		m_aFlagPos[i] = vec2(-1, -1);
	
	// get flag positions
	for(int i = 0; i < m_Collision.GetWidth()*m_Collision.GetHeight(); i++)
	{
		if(m_Collision.GetCollisionRace(i)-ENTITY_OFFSET == ENTITY_FLAGSTAND_RED)
			m_aFlagPos[TEAM_RED] = vec2((i%m_Collision.GetWidth())*32+16, (i/m_Collision.GetWidth())*32+16);
		else if(m_Collision.GetCollisionRace(i)-ENTITY_OFFSET == ENTITY_FLAGSTAND_BLUE)
			m_aFlagPos[TEAM_BLUE] = vec2((i%m_Collision.GetWidth())*32+16, (i/m_Collision.GetWidth())*32+16);
	}
}

void CGameClient::OnReset()
{
	// clear out the invalid pointers
	m_LastNewPredictedTick = -1;
	mem_zero(&g_GameClient.m_Snap, sizeof(g_GameClient.m_Snap));

	for(int i = 0; i < MAX_CLIENTS; i++)
		m_aClients[i].Reset(i);
	
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnReset();

	m_DemoSpecID = SPEC_FREEVIEW;

	// Race
	m_IsRace = false;
	m_FirstBlood = false;
	m_IsWarmupping = false;
	m_RaceMsgSent = false;
	m_ShowOthers = -1;
	for(int i = 0; i < 2; i++)
		m_aFlagPos[i] = vec2(-1, -1);
}


void CGameClient::UpdatePositions()
{
	// local character position
	if(g_Config.m_ClPredict && Client()->State() != IClient::STATE_DEMOPLAYBACK)
	{
		if(!m_Snap.m_pLocalCharacter || (m_Snap.m_pGameInfoObj && m_Snap.m_pGameInfoObj->m_GameStateFlags&GAMESTATEFLAG_GAMEOVER))
		{
			// don't use predicted
		}
		else
			m_LocalCharacterPos = mix(m_PredictedPrevChar.m_Pos, m_PredictedChar.m_Pos, Client()->PredIntraGameTick());
	}
	else if(m_Snap.m_pLocalCharacter && m_Snap.m_pLocalPrevCharacter)
	{
		m_LocalCharacterPos = mix(
			vec2(m_Snap.m_pLocalPrevCharacter->m_X, m_Snap.m_pLocalPrevCharacter->m_Y),
			vec2(m_Snap.m_pLocalCharacter->m_X, m_Snap.m_pLocalCharacter->m_Y), Client()->IntraGameTick());
	}
	
	if(!m_Snap.m_SpecInfo.m_Active && g_Config.m_GfxSpecZoom != 100)
		g_Config.m_GfxSpecZoom = 100;

	// spectator position
	if(m_Snap.m_SpecInfo.m_Active)
	{
		if(Client()->State() == IClient::STATE_DEMOPLAYBACK && DemoPlayer()->GetDemoType() == IDemoPlayer::DEMOTYPE_SERVER &&
			m_Snap.m_SpecInfo.m_SpectatorID != SPEC_FREEVIEW)
		{
			m_Snap.m_SpecInfo.m_Position = mix(
				vec2(m_Snap.m_aCharacters[m_Snap.m_SpecInfo.m_SpectatorID].m_Prev.m_X, m_Snap.m_aCharacters[m_Snap.m_SpecInfo.m_SpectatorID].m_Prev.m_Y),
				vec2(m_Snap.m_aCharacters[m_Snap.m_SpecInfo.m_SpectatorID].m_Cur.m_X, m_Snap.m_aCharacters[m_Snap.m_SpecInfo.m_SpectatorID].m_Cur.m_Y),
				Client()->IntraGameTick());
			m_Snap.m_SpecInfo.m_UsePosition = true;
		}
		else if(m_Snap.m_pSpectatorInfo && (Client()->State() == IClient::STATE_DEMOPLAYBACK || m_Snap.m_SpecInfo.m_SpectatorID != SPEC_FREEVIEW))
		{
			if(m_Snap.m_pPrevSpectatorInfo)
				m_Snap.m_SpecInfo.m_Position = mix(vec2(m_Snap.m_pPrevSpectatorInfo->m_X, m_Snap.m_pPrevSpectatorInfo->m_Y),
													vec2(m_Snap.m_pSpectatorInfo->m_X, m_Snap.m_pSpectatorInfo->m_Y), Client()->IntraGameTick());
			else
				m_Snap.m_SpecInfo.m_Position = vec2(m_Snap.m_pSpectatorInfo->m_X, m_Snap.m_pSpectatorInfo->m_Y);
			m_Snap.m_SpecInfo.m_UsePosition = true;
		}
	}
}


static void Evolve(CNetObj_Character *pCharacter, int Tick)
{
	CWorldCore TempWorld;
	CCharacterCore TempCore;
	mem_zero(&TempCore, sizeof(TempCore));
	TempCore.Init(&TempWorld, g_GameClient.Collision());
	TempCore.Read(pCharacter);
	
	while(pCharacter->m_Tick < Tick)
	{
		pCharacter->m_Tick++;
		TempCore.Tick(false);
		TempCore.Move();
		TempCore.Quantize();
	}

	TempCore.Write(pCharacter);
}


void CGameClient::OnRender()
{
	/*Graphics()->Clear(1,0,0);
	
	menus->render_background();
	return;*/
	/*
	Graphics()->Clear(1,0,0);
	Graphics()->MapScreen(0,0,100,100);
	
	Graphics()->QuadsBegin();
		Graphics()->SetColor(1,1,1,1);
		Graphics()->QuadsDraw(50, 50, 30, 30);
	Graphics()->QuadsEnd();
	
	return;*/
	
	// update the local character and spectate position
	UpdatePositions();
	
	// dispatch all input to systems
	DispatchInput();
	
	// render all systems
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnRender();
		
	// clear new tick flags
	m_NewTick = false;
	m_NewPredictedTick = false;

	// check if client info has to be resent
	if(m_LastSendInfo && Client()->State() == IClient::STATE_ONLINE && !m_pMenus->IsActive() && m_LastSendInfo+time_freq()*5 < time_get())
	{
		// resend if client info differs
		if(str_comp(g_Config.m_PlayerName, m_aClients[m_Snap.m_LocalClientID].m_aName) ||
			str_comp(g_Config.m_PlayerClan, m_aClients[m_Snap.m_LocalClientID].m_aClan) ||
			g_Config.m_PlayerCountry != m_aClients[m_Snap.m_LocalClientID].m_Country ||
			str_comp(g_Config.m_PlayerSkin, m_aClients[m_Snap.m_LocalClientID].m_aSkinName) ||
			(m_Snap.m_pGameInfoObj && !(m_Snap.m_pGameInfoObj->m_GameFlags&GAMEFLAG_TEAMS) &&	// no teamgame?
			(g_Config.m_PlayerUseCustomColor != m_aClients[m_Snap.m_LocalClientID].m_UseCustomColor ||
			g_Config.m_PlayerColorBody != m_aClients[m_Snap.m_LocalClientID].m_ColorBody ||
			g_Config.m_PlayerColorFeet != m_aClients[m_Snap.m_LocalClientID].m_ColorFeet)))
		{
			SendInfo(false);
		}
		m_LastSendInfo = 0;
	}
}

void CGameClient::OnRelease()
{
	// release all systems
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnRelease();
}

void CGameClient::OnMessage(int MsgId, CUnpacker *pUnpacker)
{
	// special messages
	if(MsgId == NETMSGTYPE_SV_EXTRAPROJECTILE)
	{
		int Num = pUnpacker->GetInt();
		
		for(int k = 0; k < Num; k++)
		{
			CNetObj_Projectile Proj;
			for(unsigned i = 0; i < sizeof(CNetObj_Projectile)/sizeof(int); i++)
				((int *)&Proj)[i] = pUnpacker->GetInt();
				
			if(pUnpacker->Error())
				return;
			
			g_GameClient.m_pItems->AddExtraProjectile(&Proj);
		}
		
		return;
	}
	else if(MsgId == NETMSGTYPE_SV_TUNEPARAMS)
	{
		// unpack the new tuning
		CTuningParams NewTuning;
		int *pParams = (int *)&NewTuning;
		for(unsigned i = 0; i < sizeof(CTuningParams)/sizeof(int); i++)
			pParams[i] = pUnpacker->GetInt();

		// check for unpacking errors
		if(pUnpacker->Error())
			return;
		
		m_ServerMode = SERVERMODE_PURE;
			
		// apply new tuning
		m_Tuning = NewTuning;
		return;
	}
	
	void *pRawMsg = m_NetObjHandler.SecureUnpackMsg(MsgId, pUnpacker);
	if(!pRawMsg)
	{
		char aBuf[256];
		str_format(aBuf, sizeof(aBuf), "dropped weird message '%s' (%d), failed on '%s'", m_NetObjHandler.GetMsgName(MsgId), MsgId, m_NetObjHandler.FailedMsgOn());
		Console()->Print(IConsole::OUTPUT_LEVEL_ADDINFO, "client", aBuf);
		return;
	}

	// TODO: this should be done smarter
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnMessage(MsgId, pRawMsg);
	
	if(MsgId == NETMSGTYPE_SV_READYTOENTER)
	{
		Client()->EnterGame();
	}
	else if (MsgId == NETMSGTYPE_SV_EMOTICON)
	{
		CNetMsg_Sv_Emoticon *pMsg = (CNetMsg_Sv_Emoticon *)pRawMsg;

		// apply
		m_aClients[pMsg->m_ClientID].m_Emoticon = pMsg->m_Emoticon;
		m_aClients[pMsg->m_ClientID].m_EmoticonStart = Client()->GameTick();
	}
	else if(MsgId == NETMSGTYPE_SV_SOUNDGLOBAL)
	{
		if(m_SuppressEvents)
			return;
		
		// don't enqueue pseudo-global sounds from demos (created by PlayAndRecord)
		CNetMsg_Sv_SoundGlobal *pMsg = (CNetMsg_Sv_SoundGlobal *)pRawMsg;
		if(pMsg->m_SoundID == SOUND_CTF_DROP || pMsg->m_SoundID == SOUND_CTF_RETURN ||
			pMsg->m_SoundID == SOUND_CTF_CAPTURE || pMsg->m_SoundID == SOUND_CTF_GRAB_EN ||
			pMsg->m_SoundID == SOUND_CTF_GRAB_PL)
			g_GameClient.m_pSounds->Enqueue(pMsg->m_SoundID);
		else
			g_GameClient.m_pSounds->Play(CSounds::CHN_GLOBAL, pMsg->m_SoundID, 1.0f, vec2(0,0));
	}
	else if(MsgId == NETMSGTYPE_SV_KILLMSG)
	{
		CNetMsg_Sv_KillMsg *pMsg = (CNetMsg_Sv_KillMsg *)pRawMsg;
		
		if(g_Config.m_ClGSound && !m_FirstBlood && m_Snap.m_LocalClientID == pMsg->m_Killer && pMsg->m_Killer != pMsg->m_Victim)
		{
			g_GameClient.m_pSounds->Play(CSounds::CHN_GUI, SOUND_SPREE_FIRSTBLOOD, 0, vec2(0,0));
			m_FirstBlood = true;
		}
	}
	else if(MsgId == NETMSGTYPE_SV_PLAYERTIME)
	{
		CNetMsg_Sv_PlayerTime *pMsg = (CNetMsg_Sv_PlayerTime *)pRawMsg;
		m_aClients[pMsg->m_ClientID].m_Score = (float)pMsg->m_Time/1000.0f;
	}
	else if(MsgId == NETMSGTYPE_SV_CHAT)
	{
		CNetMsg_Sv_Chat *pMsg = (CNetMsg_Sv_Chat *)pRawMsg;
		if(pMsg->m_ClientID == -1 && str_find(pMsg->m_pMessage, " finished in: "))
		{
			const char* pMessage = pMsg->m_pMessage;
			
			int Num = 0;
			while(str_comp_num(pMessage, " finished in: ", 14))
			{
				pMessage++;
				Num++;
				if(!pMessage[0])
					return;
			}
			
			int Minutes = 0;
			float Seconds = 0.0f;
			
			// store the name
			char Playername[MAX_NAME_LENGTH];
			str_copy(Playername, pMsg->m_pMessage, Num+1);
			
			// get time
			if(sscanf(pMessage, " finished in: %d minute(s) %f", &Minutes, &Seconds) == 2)
			{
				int PlayerID = -1;
				for(int i = 0; i < MAX_CLIENTS; i++)
					if(!str_comp(Playername, m_aClients[i].m_aName))
					{
						PlayerID = i;
						break;
					}
				
				// some proof
				if(PlayerID < 0)
					return;
				
				float Time = (float)(Minutes*60) + Seconds;
				
				if(m_aClients[PlayerID].m_Score == 0 || Time < m_aClients[PlayerID].m_Score)
					m_aClients[PlayerID].m_Score = Time;
			}
		}
	}
}

// Gamer: overall statistics
void CGameClient::RegisterStats()
{
	char Path[512];
	FILE* File;
	int AlreadyExists = 1, i;
	int LocalCid, Team = -1, Score;
	
	for(i = 0; i < Client()->SnapNumItems(IClient::SNAP_CURRENT); i++)
	{
		IClient::CSnapItem Item;
		const void *pData = Client()->SnapGetItem(IClient::SNAP_CURRENT, i, &Item);

		if(Item.m_Type == NETOBJTYPE_PLAYERINFO)
		{
			const CNetObj_PlayerInfo *pInfo = (const CNetObj_PlayerInfo *)pData;
			if(pInfo->m_Local)
			{
				Team = pInfo->m_Team;
				LocalCid = pInfo->m_ClientID;
				Score = pInfo->m_Score;
			}
		}
	}
	
	if(Team == -1) // Do not register if we're spec
		return;
	
		
	// Check if the gametype is supported
	CServerInfo CurrentServerInfo;
	Client()->GetServerInfo(&CurrentServerInfo);
	if(g_Config.m_ClRegisterStatsPure)
	{
		if(!(CurrentServerInfo.m_aGameType[0] && (!str_comp(CurrentServerInfo.m_aGameType, "DM") || 
			!str_comp(CurrentServerInfo.m_aGameType, "DM") || !str_comp(CurrentServerInfo.m_aGameType, "CTF"))))
			return;
	}
	else
	{
		if(!(CurrentServerInfo.m_aGameType[0] && (str_find(CurrentServerInfo.m_aGameType, "DM") || 
			str_find(CurrentServerInfo.m_aGameType, "DM") || str_find(CurrentServerInfo.m_aGameType, "CTF"))))
			return;
	}
	
	// Stats variables
	int games;
	
	int frags_with[NUM_WEAPONS];
	int deaths_from[NUM_WEAPONS];
	int frags;
	int deaths;
	int suicides;
	int flag_grabs;
	int flag_captures;
	int carriers_killed;
	int kills_carrying;
	int deaths_carrying;
	int won;

	int* StatsVar[] = {&games, &(frags_with[0]), &(frags_with[1]), &(frags_with[2]), &(frags_with[3]), &(frags_with[4]), &(frags_with[5]),
		&(deaths_from[0]), &(deaths_from[1]), &(deaths_from[2]), &(deaths_from[3]), &(deaths_from[4]), &(deaths_from[5]), &frags, &deaths,
		&suicides, &flag_grabs, &flag_captures, &carriers_killed, &kills_carrying, &deaths_carrying, &won};
		
	// int num_stats_var = sizeof(stats_var)/sizeof(int*);
	int NumStatsVar = 22;
	
#if defined(CONF_FAMILY_WINDOWS)
	str_format(Path, sizeof(Path), "%s\\stats.cfg", ClientUserDirectory());
#else
	str_format(Path, sizeof(Path), "%s/stats.cfg", ClientUserDirectory());
#endif

	File = fopen(Path, "r+");
	if(!File)
	{
		// Now, try to create the file
		File = fopen(Path, "w+");
		if(File)
		{
			dbg_msg("client/stats", "stats file not found, created a new one");
			AlreadyExists = 0;
		}
		else
		{
			dbg_msg("client/stats", "failed opening stats file");
			return;
		}
	}
	// -------  Reading ------
	// Loading
	if(AlreadyExists) // Don't read else
	{
		for(i = 0; i < NumStatsVar; i++)
		{
			if(!fscanf(File, "%d", StatsVar[i]))
			{
				// Corrupted file (keep what we have loaded, so that when a new version comes with more stats, we can keep the old stats.cfg)
				dbg_msg("client/stats", "file is corrupted, can't continue reading");
				break;
			}
		}
	}
	else i = 0;
	
	// Read all what we haven't loaded and turn it to 0
	for(int j = i; j < NumStatsVar; j++)
		*(StatsVar[j]) = 0;
		
	// ------ Adding our stats to the loaded stats -------
	games++;
	
	
	for(int j = 0; j < NUM_WEAPONS; j++)
	{
		frags_with[j] += m_aStats[LocalCid].m_aFragsWith[j];
		deaths_from[j] += m_aStats[LocalCid].m_aDeathsFrom[j];
	}
	frags += m_aStats[LocalCid].m_Frags;
	deaths += m_aStats[LocalCid].m_Deaths;
	suicides += m_aStats[LocalCid].m_Suicides;
	flag_grabs += m_aStats[LocalCid].m_FlagGrabs;
	flag_captures += m_aStats[LocalCid].m_FlagCaptures;
	carriers_killed += m_aStats[LocalCid].m_CarriersKilled;
	kills_carrying += m_aStats[LocalCid].m_KillsCarrying;
	deaths_carrying += m_aStats[LocalCid].m_DeathsCarrying;
	
	if(g_GameClient.m_Snap.m_pGameInfoObj && g_GameClient.m_Snap.m_pGameInfoObj->m_GameFlags&GAMEFLAG_TEAMS)
	{
		if((int)(g_GameClient.m_Snap.m_pGameDataObj->m_TeamscoreBlue > g_GameClient.m_Snap.m_pGameDataObj->m_TeamscoreRed) == Team) // If we won
			won++;
	}
	else
	{
		if(Score == g_GameClient.m_Snap.m_pGameInfoObj->m_ScoreLimit) // Let's consider won = #1
			won++;
	}
	
	// ------- Writing ------
	if(AlreadyExists) // If we haven't opened it in w+ mode yet
	{
		fclose(File);
		File = fopen(Path, "w+");
		if(!File)
			return;
	}
	
	// Now, the file is empty, write
	for(int j = 0; j < NumStatsVar; j++)
	{
		fprintf(File, "%d", *(StatsVar[j]));
		fputc('\n', File);
	}
	
	fclose(File);
	char aBuf[256];
	str_format(aBuf, sizeof(aBuf), "stats successfully saved, k/d is %d/%d", m_aStats[LocalCid].m_Frags, m_aStats[LocalCid].m_Deaths);
	Console()->Print(IConsole::OUTPUT_LEVEL_STANDARD, "client/stats", aBuf);
}

void CGameClient::OnStateChange(int NewState, int OldState)
{
	// reset everything when not already connected (to keep gathered stuff)
	if(NewState < IClient::STATE_ONLINE)
		OnReset();
	
	// then change the state
	for(int i = 0; i < m_All.m_Num; i++)
		m_All.m_paComponents[i]->OnStateChange(NewState, OldState);
}

void CGameClient::OnShutdown()
{
	m_pRaceDemo->OnShutdown();
}

void CGameClient::OnEnterGame() {}

void CGameClient::OnGameOver()
{
	if(Client()->State() != IClient::STATE_DEMOPLAYBACK)
		Client()->AutoScreenshot_Start();
	
	if(g_Config.m_ClRegisterStats)
		RegisterStats();
}

void CGameClient::OnStartGame()
{
	if(Client()->State() != IClient::STATE_DEMOPLAYBACK)
		Client()->DemoRecorder_HandleAutoStart();
}

void CGameClient::OnRconLine(const char *pLine)
{
	m_pGameConsole->PrintLine(CGameConsole::CONSOLETYPE_REMOTE, pLine);
}

void CGameClient::ProcessEvents()
{
	if(m_SuppressEvents)
		return;
	
	int SnapType = IClient::SNAP_CURRENT;
	int Num = Client()->SnapNumItems(SnapType);
	for(int Index = 0; Index < Num; Index++)
	{
		IClient::CSnapItem Item;
		const void *pData = Client()->SnapGetItem(SnapType, Index, &Item);

		if(Item.m_Type == NETEVENTTYPE_DAMAGEIND)
		{
			NETEVENT_DAMAGEIND *ev = (NETEVENT_DAMAGEIND *)pData;
			g_GameClient.m_pEffects->DamageIndicator(vec2(ev->m_X, ev->m_Y), GetDirection(ev->m_Angle));
		}
		else if(Item.m_Type == NETEVENTTYPE_EXPLOSION)
		{
			NETEVENT_EXPLOSION *ev = (NETEVENT_EXPLOSION *)pData;
			g_GameClient.m_pEffects->Explosion(vec2(ev->m_X, ev->m_Y));
		}
		else if(Item.m_Type == NETEVENTTYPE_HAMMERHIT)
		{
			NETEVENT_HAMMERHIT *ev = (NETEVENT_HAMMERHIT *)pData;
			g_GameClient.m_pEffects->HammerHit(vec2(ev->m_X, ev->m_Y));
		}
		else if(Item.m_Type == NETEVENTTYPE_SPAWN)
		{
			NETEVENT_SPAWN *ev = (NETEVENT_SPAWN *)pData;
			g_GameClient.m_pEffects->PlayerSpawn(vec2(ev->m_X, ev->m_Y));
		}
		else if(Item.m_Type == NETEVENTTYPE_DEATH)
		{
			NETEVENT_DEATH *ev = (NETEVENT_DEATH *)pData;
			g_GameClient.m_pEffects->PlayerDeath(vec2(ev->m_X, ev->m_Y), ev->m_ClientID);
		}
		else if(Item.m_Type == NETEVENTTYPE_SOUNDWORLD)
		{
			NETEVENT_SOUNDWORLD *ev = (NETEVENT_SOUNDWORLD *)pData;
			g_GameClient.m_pSounds->Play(CSounds::CHN_WORLD, ev->m_SoundID, 1.0f, vec2(ev->m_X, ev->m_Y));
		}
	}
}

void CGameClient::OnNewSnapshot()
{
	m_NewTick = true;
	
	// clear out the invalid pointers
	mem_zero(&g_GameClient.m_Snap, sizeof(g_GameClient.m_Snap));
	m_Snap.m_LocalClientID = -1;

		// mark all clients offline here
	bool Online[MAX_CLIENTS] = { 0 };
	
	// secure snapshot
	{
		int Num = Client()->SnapNumItems(IClient::SNAP_CURRENT);
		for(int Index = 0; Index < Num; Index++)
		{
			IClient::CSnapItem Item;
			void *pData = Client()->SnapGetItem(IClient::SNAP_CURRENT, Index, &Item);
			if(m_NetObjHandler.ValidateObj(Item.m_Type, pData, Item.m_DataSize) != 0)
			{
				if(g_Config.m_Debug)
				{
					char aBuf[256];
					str_format(aBuf, sizeof(aBuf), "invalidated index=%d type=%d (%s) size=%d id=%d", Index, Item.m_Type, m_NetObjHandler.GetObjName(Item.m_Type), Item.m_DataSize, Item.m_ID);
					Console()->Print(IConsole::OUTPUT_LEVEL_DEBUG, "game", aBuf);
				}
				Client()->SnapInvalidateItem(IClient::SNAP_CURRENT, Index);
			}
		}
	}
		
	ProcessEvents();

	if(g_Config.m_DbgStress)
	{
		if((Client()->GameTick()%100) == 0)
		{
			char aMessage[64];
			int MsgLen = rand()%(sizeof(aMessage)-1);
			for(int i = 0; i < MsgLen; i++)
				aMessage[i] = 'a'+(rand()%('z'-'a'));
			aMessage[MsgLen] = 0;
				
			CNetMsg_Cl_Say Msg;
			Msg.m_Team = rand()&1;
			Msg.m_pMessage = aMessage;
			Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);
		}
	}

	// go trough all the items in the snapshot and gather the info we want
	{
		m_Snap.m_aTeamSize[TEAM_RED] = m_Snap.m_aTeamSize[TEAM_BLUE] = 0;
		
		// TeeComp
		for(int i = 0; i < MAX_CLIENTS; i++)
			m_aStats[i].m_Active = false;

		int Num = Client()->SnapNumItems(IClient::SNAP_CURRENT);
		for(int i = 0; i < Num; i++)
		{
			IClient::CSnapItem Item;
			const void *pData = Client()->SnapGetItem(IClient::SNAP_CURRENT, i, &Item);

			if(Item.m_Type == NETOBJTYPE_CLIENTINFO)
			{
				const CNetObj_ClientInfo *pInfo = (const CNetObj_ClientInfo *)pData;
				int ClientID = Item.m_ID;
				IntsToStr(&pInfo->m_Name0, 4, m_aClients[ClientID].m_aName);
				IntsToStr(&pInfo->m_Clan0, 3, m_aClients[ClientID].m_aClan);
				m_aClients[ClientID].m_Country = GetCountryIndex(pInfo->m_Country);
				IntsToStr(&pInfo->m_Skin0, 6, m_aClients[ClientID].m_aSkinName);
				
				m_aClients[ClientID].m_UseCustomColor = pInfo->m_UseCustomColor;
				m_aClients[ClientID].m_ColorBody = pInfo->m_ColorBody;
				m_aClients[ClientID].m_ColorFeet = pInfo->m_ColorFeet;
				
				// prepare the info
				if(m_aClients[ClientID].m_aSkinName[0] == 'x' || m_aClients[ClientID].m_aSkinName[1] == '_')
					str_copy(m_aClients[ClientID].m_aSkinName, "default", 64);
					
				m_aClients[ClientID].m_SkinInfo.m_ColorBody = m_pSkins->GetColorV4(m_aClients[ClientID].m_ColorBody);
				m_aClients[ClientID].m_SkinInfo.m_ColorFeet = m_pSkins->GetColorV4(m_aClients[ClientID].m_ColorFeet);
				m_aClients[ClientID].m_SkinInfo.m_Size = 64;
				
				// find new skin
				m_aClients[ClientID].m_SkinID = g_GameClient.m_pSkins->Find(m_aClients[ClientID].m_aSkinName);
				if(m_aClients[ClientID].m_SkinID < 0)
				{
					m_aClients[ClientID].m_SkinID = g_GameClient.m_pSkins->Find("default");
					if(m_aClients[ClientID].m_SkinID < 0)
						m_aClients[ClientID].m_SkinID = 0;
				}
				
				if(m_aClients[ClientID].m_UseCustomColor)
					m_aClients[ClientID].m_SkinInfo.m_Texture = g_GameClient.m_pSkins->Get(m_aClients[ClientID].m_SkinID)->m_ColorTexture;
				else
				{
					m_aClients[ClientID].m_SkinInfo.m_Texture = g_GameClient.m_pSkins->Get(m_aClients[ClientID].m_SkinID)->m_OrgTexture;
					m_aClients[ClientID].m_SkinInfo.m_ColorBody = vec4(1,1,1,1);
					m_aClients[ClientID].m_SkinInfo.m_ColorFeet = vec4(1,1,1,1);
				}

				m_aClients[ClientID].UpdateRenderInfo(ClientID);

				// mark Player as online
				Online[ClientID] = true;
			}
			else if(Item.m_Type == NETOBJTYPE_PLAYERINFO)
			{
				const CNetObj_PlayerInfo *pInfo = (const CNetObj_PlayerInfo *)pData;
				
				m_aClients[pInfo->m_ClientID].m_Team = pInfo->m_Team;
				m_aClients[pInfo->m_ClientID].m_Active = true;
				m_Snap.m_paPlayerInfos[pInfo->m_ClientID] = pInfo;
				m_Snap.m_NumPlayers++;
				
				if(pInfo->m_Local)
				{
					m_Snap.m_LocalClientID = Item.m_ID;
					m_Snap.m_pLocalInfo = pInfo;
					
					if(pInfo->m_Team == TEAM_SPECTATORS)
					{
						m_Snap.m_SpecInfo.m_Active = true;
						m_Snap.m_SpecInfo.m_SpectatorID = SPEC_FREEVIEW;
					}
				}
				
				// calculate team-balance
				if(pInfo->m_Team != TEAM_SPECTATORS)
				{
					m_Snap.m_aTeamSize[pInfo->m_Team]++;
					m_aStats[pInfo->m_ClientID].m_Active = true;
				}
				
			}
			else if(Item.m_Type == NETOBJTYPE_CHARACTER)
			{
				const void *pOld = Client()->SnapFindItem(IClient::SNAP_PREV, NETOBJTYPE_CHARACTER, Item.m_ID);
				m_Snap.m_aCharacters[Item.m_ID].m_Cur = *((const CNetObj_Character *)pData);
				if(pOld)
				{
					m_Snap.m_aCharacters[Item.m_ID].m_Active = true;
					m_Snap.m_aCharacters[Item.m_ID].m_Prev = *((const CNetObj_Character *)pOld);

					if(m_Snap.m_aCharacters[Item.m_ID].m_Prev.m_Tick)
						Evolve(&m_Snap.m_aCharacters[Item.m_ID].m_Prev, Client()->PrevGameTick());
					if(m_Snap.m_aCharacters[Item.m_ID].m_Cur.m_Tick)
						Evolve(&m_Snap.m_aCharacters[Item.m_ID].m_Cur, Client()->GameTick());
				}
			}
			else if(Item.m_Type == NETOBJTYPE_SPECTATORINFO)
			{
				m_Snap.m_pSpectatorInfo = (const CNetObj_SpectatorInfo *)pData;
				m_Snap.m_pPrevSpectatorInfo = (const CNetObj_SpectatorInfo *)Client()->SnapFindItem(IClient::SNAP_PREV, NETOBJTYPE_SPECTATORINFO, Item.m_ID);

				m_Snap.m_SpecInfo.m_SpectatorID = m_Snap.m_pSpectatorInfo->m_SpectatorID;
			}
			else if(Item.m_Type == NETOBJTYPE_GAMEINFO)
			{
				// static bool s_GameOver = 0;
				m_Snap.m_pGameInfoObj = (const CNetObj_GameInfo *)pData;
				if(!m_LastGameOver && m_Snap.m_pGameInfoObj->m_GameStateFlags&GAMESTATEFLAG_GAMEOVER)
					OnGameOver();
				else if(m_LastGameOver && !m_Snap.m_pGameInfoObj->m_GameStateFlags&GAMESTATEFLAG_GAMEOVER)
					OnGameRestart();
				if(m_LastRoundStartTick != m_Snap.m_pGameInfoObj->m_RoundStartTick)
					OnRoundStart();
				
				m_LastGameOver = m_Snap.m_pGameInfoObj->m_GameStateFlags&GAMESTATEFLAG_GAMEOVER;
				m_LastRoundStartTick = m_Snap.m_pGameInfoObj->m_RoundStartTick;
			}
			else if(Item.m_Type == NETOBJTYPE_GAMEDATA)
			{
				// m_Snap.m_pGameDataObj = (const CNetObj_GameData *)pData;
				static int s_FlagCarrierRed = FLAG_ATSTAND;
				static int s_FlagCarrierBlue = FLAG_ATSTAND;
				const CNetObj_GameData *pGameData = (const CNetObj_GameData *)pData;
				
				m_Snap.m_pGameDataObj = pGameData;
 				m_Snap.m_GameDataSnapID = Item.m_ID;
				
				if(s_FlagCarrierRed == FLAG_ATSTAND && pGameData->m_FlagCarrierRed >= 0)
					OnFlagGrab(TEAM_RED);
				else if(s_FlagCarrierBlue == FLAG_ATSTAND && pGameData->m_FlagCarrierBlue >= 0)
					OnFlagGrab(TEAM_BLUE);

				s_FlagCarrierRed = pGameData->m_FlagCarrierRed;
				s_FlagCarrierBlue = pGameData->m_FlagCarrierBlue;
				m_Snap.m_GameDataSnapID = Item.m_ID;
			}
			else if(Item.m_Type == NETOBJTYPE_FLAG)
				m_Snap.m_paFlags[Item.m_ID%2] = (const CNetObj_Flag *)pData;
		}

		// TeeComp
		for(int i = 0; i < MAX_CLIENTS; i++)
		{
			if(m_aStats[i].m_Active && !m_aStats[i].m_WasActive)
			{
				m_aStats[i].Reset(); // Client connected, reset stats.
				m_aStats[i].m_Active = true;
				m_aStats[i].m_JoinDate = Client()->GameTick();
			}
			m_aStats[i].m_WasActive = m_aStats[i].m_Active;
		}
	}
	
	// setup local pointers
	if(m_Snap.m_LocalClientID >= 0)
	{
		CSnapState::CCharacterInfo *c = &m_Snap.m_aCharacters[m_Snap.m_LocalClientID];
		if(c->m_Active)
		{
			m_Snap.m_pLocalCharacter = &c->m_Cur;
			m_Snap.m_pLocalPrevCharacter = &c->m_Prev;
			m_LocalCharacterPos = vec2(m_Snap.m_pLocalCharacter->m_X, m_Snap.m_pLocalCharacter->m_Y);
		}
		else if(Client()->SnapFindItem(IClient::SNAP_PREV, NETOBJTYPE_CHARACTER, m_Snap.m_LocalClientID))
		{
			// player died
			m_pControls->OnPlayerDeath();
		}
	}
	else
	{
		m_Snap.m_SpecInfo.m_Active = true;
		if(Client()->State() == IClient::STATE_DEMOPLAYBACK && DemoPlayer()->GetDemoType() == IDemoPlayer::DEMOTYPE_SERVER &&
			m_DemoSpecID != SPEC_FREEVIEW && m_Snap.m_aCharacters[m_DemoSpecID].m_Active)
			m_Snap.m_SpecInfo.m_SpectatorID = m_DemoSpecID;
		else
			m_Snap.m_SpecInfo.m_SpectatorID = SPEC_FREEVIEW;
	}

	// clear out unneeded client data
	for(int i = 0; i < MAX_CLIENTS; ++i)
	{
		if(!m_Snap.m_paPlayerInfos[i] && m_aClients[i].m_Active)
			m_aClients[i].Reset(i);
	}

	// update friend state
	for(int i = 0; i < MAX_CLIENTS; ++i)
	{
		if(i == m_Snap.m_LocalClientID || !m_Snap.m_paPlayerInfos[i] || !Friends()->IsFriend(m_aClients[i].m_aName, m_aClients[i].m_aClan, true))
			m_aClients[i].m_Friend = false;
		else
			m_aClients[i].m_Friend = true;
	}

	// sort player infos by score
	mem_copy(m_Snap.m_paInfoByScore, m_Snap.m_paPlayerInfos, sizeof(m_Snap.m_paInfoByScore));
	for(int k = 0; k < MAX_CLIENTS-1; k++) // ffs, bubblesort
	{
		for(int i = 0; i < MAX_CLIENTS-k-1; i++)
		{
			if(m_IsRace)
			{
				if(m_Snap.m_paInfoByScore[i+1] && (!m_Snap.m_paInfoByScore[i] || m_aClients[m_Snap.m_paInfoByScore[i]->m_ClientID].m_Score == 0 || (m_aClients[m_Snap.m_paInfoByScore[i]->m_ClientID].m_Score > m_aClients[m_Snap.m_paInfoByScore[i+1]->m_ClientID].m_Score && m_aClients[m_Snap.m_paInfoByScore[i+1]->m_ClientID].m_Score != 0)))
				{
					const CNetObj_PlayerInfo *pTmp = m_Snap.m_paInfoByScore[i];
					m_Snap.m_paInfoByScore[i] = m_Snap.m_paInfoByScore[i+1];
					m_Snap.m_paInfoByScore[i+1] = pTmp;
				}
			}
			else
			{
				if(m_Snap.m_paInfoByScore[i+1] && (!m_Snap.m_paInfoByScore[i] || m_Snap.m_paInfoByScore[i]->m_Score < m_Snap.m_paInfoByScore[i+1]->m_Score))
				{
					const CNetObj_PlayerInfo *pTmp = m_Snap.m_paInfoByScore[i];
					m_Snap.m_paInfoByScore[i] = m_Snap.m_paInfoByScore[i+1];
					m_Snap.m_paInfoByScore[i+1] = pTmp;
				}
			}
		}
	}
	
	CTuningParams StandardTuning;
	StandardTuning.m_PlayerCollision = 1;
	StandardTuning.m_PlayerHooking = 1;
	CServerInfo CurrentServerInfo;
	Client()->GetServerInfo(&CurrentServerInfo);
	if(CurrentServerInfo.m_aGameType[0] != '0')
	{
		if(str_comp(CurrentServerInfo.m_aGameType, "DM") != 0 && str_comp(CurrentServerInfo.m_aGameType, "TDM") != 0 && str_comp(CurrentServerInfo.m_aGameType, "CTF") != 0)
			m_ServerMode = SERVERMODE_MOD;
		else if(mem_comp(&StandardTuning, &m_Tuning, sizeof(CTuningParams)) == 0)
			m_ServerMode = SERVERMODE_PURE;
		else
			m_ServerMode = SERVERMODE_PUREMOD;
	}
	// send race msg
	if(m_Snap.m_pLocalInfo)
	{
		CServerInfo CurrentServerInfo;
		Client()->GetServerInfo(&CurrentServerInfo);
		if(str_find_nocase(CurrentServerInfo.m_aGameType, "race") || str_find_nocase(CurrentServerInfo.m_aGameType, "fastcap"))
		{
			if(!m_IsRace)
			{
				m_IsRace = true;
				
				// send login
				if(g_Config.m_ClApiToken[0])
				{
					char aLogin[64];
					str_format(aLogin, sizeof(aLogin), "teerace:%s", g_Config.m_ClApiToken);
					Client()->RconAuth("", aLogin);
				}
			}
			
			if(str_find_nocase(CurrentServerInfo.m_aGameType, "fastcap"))
			{
				// dbg_msg("client", "THIS IS... FASTCAP!");
				m_IsFastCap = true;
			}
			
			if(!m_RaceMsgSent && m_Snap.m_pLocalInfo)
			{
				CNetMsg_Cl_IsRace Msg;
				Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);
				m_RaceMsgSent = true;
			}
			
			if(m_ShowOthers == -1 || (m_ShowOthers > -1 && m_ShowOthers != g_Config.m_ClShowOthers))
			{
				if(m_ShowOthers == -1 && g_Config.m_ClShowOthers)
					m_ShowOthers = 1;
				else
				{
					CNetMsg_Cl_RaceShowOthers Msg;
					Msg.m_Active = g_Config.m_ClShowOthers;
					Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);
					
					m_ShowOthers = g_Config.m_ClShowOthers;
				}
			}
		}
	}
	
	// reset all scores of offline players
	if(m_IsRace)
	{
		for(int i = 0; i < MAX_CLIENTS; i++)
		{
			if(!Online[i])
				m_aClients[i].m_Score = 0.0f;
		}
	}
	// update render info
	for(int i = 0; i < MAX_CLIENTS; i++)
		m_aClients[i].UpdateRenderInfo(i);
}

void CGameClient::OnPredict()
{
	// store the previous values so we can detect prediction errors
	CCharacterCore BeforePrevChar = m_PredictedPrevChar;
	CCharacterCore BeforeChar = m_PredictedChar;

	// we can't predict without our own id or own character
	if(m_Snap.m_LocalClientID == -1 || !m_Snap.m_aCharacters[m_Snap.m_LocalClientID].m_Active)
		return;
	
	// don't predict anything if we are paused
	if(m_Snap.m_pGameInfoObj && m_Snap.m_pGameInfoObj->m_GameStateFlags&GAMESTATEFLAG_PAUSED)
	{
		if(m_Snap.m_pLocalCharacter)
			m_PredictedChar.Read(m_Snap.m_pLocalCharacter);
		if(m_Snap.m_pLocalPrevCharacter)
			m_PredictedPrevChar.Read(m_Snap.m_pLocalPrevCharacter);
		return;
	}

	// repredict character
	CWorldCore World;
	World.m_Tuning = m_Tuning;

	// search for players
	for(int i = 0; i < MAX_CLIENTS; i++)
	{
		if(!m_Snap.m_aCharacters[i].m_Active)
			continue;
			
		g_GameClient.m_aClients[i].m_Predicted.Init(&World, Collision());
		World.m_apCharacters[i] = &g_GameClient.m_aClients[i].m_Predicted;
		g_GameClient.m_aClients[i].m_Predicted.Read(&m_Snap.m_aCharacters[i].m_Cur);
	}
	
	// predict
	for(int Tick = Client()->GameTick()+1; Tick <= Client()->PredGameTick(); Tick++)
	{
		// fetch the local
		if(Tick == Client()->PredGameTick() && World.m_apCharacters[m_Snap.m_LocalClientID])
			m_PredictedPrevChar = *World.m_apCharacters[m_Snap.m_LocalClientID];
		
		// first calculate where everyone should move
		for(int c = 0; c < MAX_CLIENTS; c++)
		{
			if(!World.m_apCharacters[c])
				continue;

			mem_zero(&World.m_apCharacters[c]->m_Input, sizeof(World.m_apCharacters[c]->m_Input));
			if(m_Snap.m_LocalClientID == c)
			{
				// apply player input
				int *pInput = Client()->GetInput(Tick);
				if(pInput)
					World.m_apCharacters[c]->m_Input = *((CNetObj_PlayerInput*)pInput);
				World.m_apCharacters[c]->Tick(true);
			}
			else
				World.m_apCharacters[c]->Tick(false);

		}

		// move all players and quantize their data
		for(int c = 0; c < MAX_CLIENTS; c++)
		{
			if(!World.m_apCharacters[c])
				continue;

			World.m_apCharacters[c]->Move();
			World.m_apCharacters[c]->Quantize();
		}
		
		// check if we want to trigger effects
		if(Tick > m_LastNewPredictedTick)
		{
			m_LastNewPredictedTick = Tick;
			m_NewPredictedTick = true;
			
			if(m_Snap.m_LocalClientID != -1 && World.m_apCharacters[m_Snap.m_LocalClientID])
			{
				vec2 Pos = World.m_apCharacters[m_Snap.m_LocalClientID]->m_Pos;
				int Events = World.m_apCharacters[m_Snap.m_LocalClientID]->m_TriggeredEvents;
				if(Events&COREEVENT_GROUND_JUMP) g_GameClient.m_pSounds->PlayAndRecord(CSounds::CHN_WORLD, SOUND_PLAYER_JUMP, 1.0f, Pos);
				
				/*if(events&COREEVENT_AIR_JUMP)
				{
					GameClient.effects->air_jump(pos);
					GameClient.sounds->play_and_record(SOUNDS::CHN_WORLD, SOUND_PLAYER_AIRJUMP, 1.0f, pos);
				}*/
				
				//if(events&COREEVENT_HOOK_LAUNCH) snd_play_random(CHN_WORLD, SOUND_HOOK_LOOP, 1.0f, pos);
				//if(events&COREEVENT_HOOK_ATTACH_PLAYER) snd_play_random(CHN_WORLD, SOUND_HOOK_ATTACH_PLAYER, 1.0f, pos);
				if(Events&COREEVENT_HOOK_ATTACH_GROUND) g_GameClient.m_pSounds->PlayAndRecord(CSounds::CHN_WORLD, SOUND_HOOK_ATTACH_GROUND, 1.0f, Pos);
				if(Events&COREEVENT_HOOK_HIT_NOHOOK) g_GameClient.m_pSounds->PlayAndRecord(CSounds::CHN_WORLD, SOUND_HOOK_NOATTACH, 1.0f, Pos);
				//if(events&COREEVENT_HOOK_RETRACT) snd_play_random(CHN_WORLD, SOUND_PLAYER_JUMP, 1.0f, pos);
			}
		}
		
		if(Tick == Client()->PredGameTick() && World.m_apCharacters[m_Snap.m_LocalClientID])
			m_PredictedChar = *World.m_apCharacters[m_Snap.m_LocalClientID];
	}
	
	if(g_Config.m_Debug && g_Config.m_ClPredict && m_PredictedTick == Client()->PredGameTick())
	{
		CNetObj_CharacterCore Before = {0}, Now = {0}, BeforePrev = {0}, NowPrev = {0};
		BeforeChar.Write(&Before);
		BeforePrevChar.Write(&BeforePrev);
		m_PredictedChar.Write(&Now);
		m_PredictedPrevChar.Write(&NowPrev);

		if(mem_comp(&Before, &Now, sizeof(CNetObj_CharacterCore)) != 0)
		{
			Console()->Print(IConsole::OUTPUT_LEVEL_DEBUG, "client", "prediction error");
			for(unsigned i = 0; i < sizeof(CNetObj_CharacterCore)/sizeof(int); i++)
				if(((int *)&Before)[i] != ((int *)&Now)[i])
				{
					char aBuf[256];
					str_format(aBuf, sizeof(aBuf), "    %d %d %d  (%d %d)", i, ((int *)&Before)[i], ((int *)&Now)[i], ((int *)&BeforePrev)[i], ((int *)&NowPrev)[i]);
					Console()->Print(IConsole::OUTPUT_LEVEL_DEBUG, "client", aBuf);
				}
		}
	}
	
	m_PredictedTick = Client()->PredGameTick();
}

// Teecomp
void CGameClient::OnGameRestart()
{	
	m_pTeecompStats->OnReset();
}

void CGameClient::OnActivateEditor()
{
	OnRelease();
}
void CGameClient::OnRoundStart()
{
	for(int i=0; i<MAX_CLIENTS; i++)
		m_aStats[i].Reset();
}

void CGameClient::OnFlagGrab(int ID)
{
	if(ID == TEAM_RED)
		m_aStats[m_Snap.m_pGameDataObj->m_FlagCarrierRed].m_FlagGrabs++;
	else
		m_aStats[m_Snap.m_pGameDataObj->m_FlagCarrierBlue].m_FlagGrabs++;
}

CGameClient::CClientStats::CClientStats()
{
	m_JoinDate  = 0;
	m_Active    = false;
	m_WasActive = false;
	m_Frags     = 0;
	m_Deaths    = 0;
	m_Suicides  = 0;
	for(int j = 0; j < NUM_WEAPONS; j++)
	{
		m_aFragsWith[j]  = 0;
		m_aDeathsFrom[j] = 0;
	}
	m_FlagGrabs      = 0;
	m_FlagCaptures   = 0;
	m_CarriersKilled = 0;
	m_KillsCarrying  = 0;
	m_DeathsCarrying = 0;
}

void CGameClient::CClientStats::Reset()
{
	m_JoinDate  = 0;
	m_Active    = false;
	m_WasActive = false;
	m_Frags     = 0;
	m_Deaths    = 0;
	m_Suicides  = 0;
	m_BestSpree = 0;
	m_CurrentSpree = 0;
	m_CurrentHumiliation = 0;
	for(int j = 0; j < NUM_WEAPONS; j++)
	{
		m_aFragsWith[j]  = 0;
		m_aDeathsFrom[j] = 0;
	}
	m_FlagGrabs      = 0;
	m_FlagCaptures   = 0;
	m_CarriersKilled = 0;
	m_KillsCarrying  = 0;
	m_DeathsCarrying = 0;
}

void CGameClient::CClientData::UpdateRenderInfo(int ClientID)
{
	m_RenderInfo = m_SkinInfo;

	// force team colors
	if(g_GameClient.m_Snap.m_pGameInfoObj && g_GameClient.m_Snap.m_pGameInfoObj->m_GameFlags&GAMEFLAG_TEAMS)
	{
		int LocalTeam;
		if(g_GameClient.m_Snap.m_pLocalInfo)
			LocalTeam = g_GameClient.m_Snap.m_pLocalInfo->m_Team;
		else // local_info null when joining a server
			LocalTeam = 0;
		if(m_Team != TEAM_SPECTATORS)
		{			
			const char* pForcedSkin;
			int Sid = m_SkinID;
			if(ClientID != g_GameClient.m_Snap.m_LocalClientID && CTeecompUtils::GetForcedSkinName(m_Team, LocalTeam, pForcedSkin))
				Sid = max(0, g_GameClient.m_pSkins->Find(pForcedSkin));

			if(CTeecompUtils::GetForceDmColors(m_Team, LocalTeam))
			{
				m_RenderInfo.m_Texture = g_GameClient.m_pSkins->Get(Sid)->m_OrgTexture;
				m_RenderInfo.m_ColorBody = vec4(1,1,1,1);
				m_RenderInfo.m_ColorFeet = vec4(1,1,1,1);
			}
			else
			{
				m_RenderInfo.m_Texture = g_GameClient.m_pSkins->Get(Sid)->m_ColorTexture;
				vec3 Col = CTeecompUtils::GetTeamColor(m_Team, LocalTeam, g_Config.m_TcColoredTeesTeam1,
					g_Config.m_TcColoredTeesTeam2, g_Config.m_TcColoredTeesMethod);
				m_RenderInfo.m_ColorBody = vec4(Col.r, Col.g, Col.b, 1.0f);
				m_RenderInfo.m_ColorFeet = vec4(Col.r, Col.g, Col.b, 1.0f);
			}
		}
		else
		{
			m_RenderInfo.m_ColorBody = g_GameClient.m_pSkins->GetColorV4(12895054);
			m_RenderInfo.m_ColorFeet = g_GameClient.m_pSkins->GetColorV4(12895054);
		}	
	}
	else if(g_Config.m_TcForceSkinTeam1 && ClientID != g_GameClient.m_Snap.m_LocalClientID) // Force DM skin
	{
		const CSkins::CSkin* pSkin;
		pSkin = g_GameClient.m_pSkins->Get(max(0, g_GameClient.m_pSkins->Find(g_Config.m_TcForcedSkin1)));
		if(m_UseCustomColor)
			m_RenderInfo.m_Texture = pSkin->m_ColorTexture;
		else
			m_RenderInfo.m_Texture = pSkin->m_OrgTexture;
	}		
}

void CGameClient::CClientData::Reset(int ClientID)
{
	m_aName[0] = 0;
	m_aClan[0] = 0;
	m_Country = -1;
	m_SkinID = 0;
	m_Team = 0;
	m_Angle = 0;
	m_Emoticon = 0;
	m_EmoticonStart = -1;
	m_Active = false;
	m_ChatIgnore = false;
	m_Score = 0.0f;
	m_SkinInfo.m_Texture = g_GameClient.m_pSkins->Get(0)->m_ColorTexture;
	m_SkinInfo.m_ColorBody = vec4(1,1,1,1);
	m_SkinInfo.m_ColorFeet = vec4(1,1,1,1);
	UpdateRenderInfo(ClientID);
}

void CGameClient::SendSwitchTeam(int Team)
{
	CNetMsg_Cl_SetTeam Msg;
	Msg.m_Team = Team;
	Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);	
}

void CGameClient::SendInfo(bool Start)
{
	if(Start)
	{	
		char Clan[64];
		CNetMsg_Cl_StartInfo Msg;
		Msg.m_pName = g_Config.m_PlayerName;
		if(g_Config.m_ClSendClientInfo && !str_length(g_Config.m_PlayerClan))
		{
			Clan[0] = '.';
			Clan[1] = ' ';
			Clan[2] = 0;
			Msg.m_pClan = Clan;
		}
		else if(g_Config.m_ClSendClientInfo && str_length(g_Config.m_PlayerClan) <= 12)
		{
			str_copy(Clan, g_Config.m_PlayerClan, sizeof(Clan));
			
			if(str_length(g_Config.m_PlayerClan) == 12)
				Clan[str_length(g_Config.m_PlayerClan)-1] = ' ';
			else
				Clan[str_length(g_Config.m_PlayerClan)] = ' ';
				
			Msg.m_pClan = Clan;
		}
		else
			Msg.m_pClan = g_Config.m_PlayerClan;
		Msg.m_Country = g_Config.m_PlayerCountry;
		Msg.m_pSkin = g_Config.m_PlayerSkin;
		Msg.m_UseCustomColor = g_Config.m_PlayerUseCustomColor;
		Msg.m_ColorBody = g_Config.m_PlayerColorBody;
		Msg.m_ColorFeet = g_Config.m_PlayerColorFeet;
		Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);	
	}
	else
	{
		char Clan[64];
		CNetMsg_Cl_ChangeInfo Msg;
		Msg.m_pName = g_Config.m_PlayerName;
		if(g_Config.m_ClSendClientInfo && !str_length(g_Config.m_PlayerClan))
		{
			Clan[0] = '.';
			Clan[1] = ' ';
			Clan[2] = 0;
			Msg.m_pClan = Clan;
		}
		else if(g_Config.m_ClSendClientInfo && str_length(g_Config.m_PlayerClan) <= 12)
		{
			str_copy(Clan, g_Config.m_PlayerClan, sizeof(Clan));
			
			if(str_length(g_Config.m_PlayerClan) == 12)
				Clan[str_length(g_Config.m_PlayerClan)-1] = ' '; // non-printable character
			else
				Clan[str_length(g_Config.m_PlayerClan)] = ' ';
				
			Msg.m_pClan = Clan;
		}
		else
			Msg.m_pClan = g_Config.m_PlayerClan;
		Msg.m_Country = g_Config.m_PlayerCountry;
		Msg.m_pSkin = g_Config.m_PlayerSkin;
		Msg.m_UseCustomColor = g_Config.m_PlayerUseCustomColor;
		Msg.m_ColorBody = g_Config.m_PlayerColorBody;
		Msg.m_ColorFeet = g_Config.m_PlayerColorFeet;
		Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);

		// activate timer to resend the info if it gets filtered
		if(!m_LastSendInfo || m_LastSendInfo+time_freq()*5 < time_get())
			m_LastSendInfo = time_get();
	}
}

void CGameClient::SendKill(int ClientID)
{
	CNetMsg_Cl_Kill Msg;
	Client()->SendPackMsg(&Msg, MSGFLAG_VITAL);	
}

void CGameClient::ConTeam(IConsole::IResult *pResult, void *pUserData)
{
	((CGameClient*)pUserData)->SendSwitchTeam(pResult->GetInteger(0));
}

void CGameClient::ConKill(IConsole::IResult *pResult, void *pUserData)
{
	((CGameClient*)pUserData)->SendKill(-1);
}

void CGameClient::ConMinimap(IConsole::IResult *pResult, void *pUserData)
{
	g_Config.m_ShowMM = pResult->GetInteger(0) != 0;
}

// Put this in .h
struct AREA
{
	float XBegin;
	float YBegin;
	float XEnd;
	float YEnd;
};
struct MAPINFO
{
	char mapname[32];
	AREA red;
	AREA blue;
};

AREA Area(int XBegin, int YBegin, int XEnd, int YEnd)
{
	AREA m;
	const float TileSize = 32.0f;
	m.XBegin = (float)(XBegin < 0 ? XBegin : TileSize*(float)XBegin);
	m.XEnd = (float)(XEnd < 0 ? XEnd : TileSize*(float)XEnd);
	m.YBegin = (float)(YBegin < 0 ? YBegin : TileSize*(float)YBegin);
	m.YEnd = (float)(YEnd < 0 ? YEnd : TileSize*(float)YEnd);
	return m;
}

MAPINFO MapInfo(char mapname[32], AREA red, AREA blue)
{
	MAPINFO m;
	str_copy(m.mapname, mapname, sizeof(m.mapname));
	m.red = red;
	m.blue = blue;
	return m;
}

bool IntoArea(vec2 p, AREA a)
{
	const int INF = -1;
	
	// dbg_msg("client", "p.x=%.0f p.y=%.0f, a=%.0f:%.0f:%.0f:%.0f", p.x, p.y, a.XBegin, a.YBegin, a.XEnd, a.YEnd);
	
	if(a.XBegin != INF && p.x < a.XBegin)
		return false;
	if(a.YBegin != INF && p.y < a.YBegin)
		return false;
	if(a.XEnd != INF && p.x > a.XEnd)
		return false;
	if(a.YEnd != INF && p.y > a.YEnd)
		return false;
	return true;
}

void CGameClient::ConSendPosition(IConsole::IResult *pResult, void *pUserData)
{
	// if(!(g_GameClient.m_Snap.m_aCharacters[g_GameClient.m_Snap.m_LocalClientID]))
	if(!g_GameClient.m_Snap.m_aCharacters[g_GameClient.m_Snap.m_LocalClientID].m_Active)
		return;
	
	dbg_msg("client", "called send position");
		
	vec2 pos = vec2(g_GameClient.m_Snap.m_aCharacters[g_GameClient.m_Snap.m_LocalClientID].m_Cur.m_X,
		g_GameClient.m_Snap.m_aCharacters[g_GameClient.m_Snap.m_LocalClientID].m_Cur.m_Y);
		
	CServerInfo CurrentServerInfo;
	((CGameClient*)pUserData)->Client()->GetServerInfo(&CurrentServerInfo);
	
	MAPINFO Infos[9];
	int i = 0;
	const int INF = -1;
	Infos[i++] = MapInfo("ctf1", Area(INF,INF,32,26), Area(32,INF,INF,26));
	Infos[i++] = MapInfo("ctf2", Area(INF,INF,55,56), Area(220-65,INF,INF,56));
	Infos[i++] = MapInfo("ctf3", Area(INF,INF,40,INF), Area(103,INF,INF,INF));
	Infos[i++] = MapInfo("ctf4", Area(INF,INF,52,INF), Area(130,INF,INF,INF));
	Infos[i++] = MapInfo("ctf5", Area(INF,INF,62,INF), Area(208,INF,INF,INF));
	Infos[i++] = MapInfo("ctf6", Area(INF,INF,63,INF), Area(214,INF,INF,INF));
	Infos[i++] = MapInfo("ctf7", Area(INF,INF,32,24), Area(32,INF,INF,24));
	Infos[i++] = MapInfo("ctf4_old", Area(INF,INF,44,INF), Area(155,INF,INF,INF));
	Infos[i++] = MapInfo("ctf_classic4", Area(INF,INF,44,INF), Area(155,INF,INF,INF));
		
	
	char MapName[32];
	str_copy(MapName, CurrentServerInfo.m_aMap, sizeof(MapName));
	for(int j = 0; j < i; j++)
	{
		if(!str_comp(MapName, Infos[j].mapname))
		{
			// Check for Red base
			if(IntoArea(pos, Infos[j].red))
				((CGameClient*)pUserData)->m_pChat->Say(1, g_Config.m_ClRedbaseMsg); // Change that with configs
			
			// Check for Blue base
			else if(IntoArea(pos, Infos[j].blue))
				((CGameClient*)pUserData)->m_pChat->Say(1, g_Config.m_ClBluebaseMsg); // Change that with configs
			
			// Then it's middle
			else
				((CGameClient*)pUserData)->m_pChat->Say(1, g_Config.m_ClMiddleMsg); // Change that with configs
		}
	}
}

void CGameClient::ConchainSpecialInfoupdate(IConsole::IResult *pResult, void *pUserData, IConsole::FCommandCallback pfnCallback, void *pCallbackUserData)
{
	pfnCallback(pResult, pCallbackUserData);
	if(pResult->NumArguments())
		((CGameClient*)pUserData)->SendInfo(false);
}

IGameClient *CreateGameClient()
{
	return &g_GameClient;
}
